/* 
 * utility to print the value of _stksize from gcc-cc1.ttp
 *
 *	Usage: printstk [<filename>]
 *		if <filename> is not specified defaults to .\gcc-cc1.ttp
 *	++jrb
 *
 *      modified to print a value of _initial_stack in cases when this
 *      is defined instead of _stksize -- mj
 */

#include <stdio.h>
#include <stdlib.h>
#include <unixlib.h>
#include <string.h>
#include <st-out.h>

long lseek(int, long, int);
static char *sym_names[] = { "__stksize", "__initial_stack" };

long find_offset (fd, fn, what)
int fd;
char *fn;
int *what;
{
    struct aexec head;
    struct asym  sym;
    int found;
    int	all = 0;
    int index = 1;
    
    if(read(fd, &head, sizeof(head)) != sizeof(head))
    {
	perror(fn);
	exit(2);
    }
    if(head.a_magic != CMAGIC)
    {
	fprintf(stderr,"Invalid magic number %x\n", head.a_magic);
	exit(3);
    }
    if(head.a_syms == 0)
    {
	fprintf(stderr,"%s: no symbol table\n", fn);
	exit(4);
    }
    if(lseek(fd, head.a_text+head.a_data, 1) != 
       (head.a_text+head.a_data+sizeof(head)))
    {
	perror(fn);
	exit(5);
    }
    for(;;)
    {
	if(index && (read(fd, &sym, sizeof(sym)) != sizeof(sym)))
	{
	    fprintf(stderr, "symbol _stksize not found\n");
	    exit(6);
	}
	/* after symbol read check first for _stksize */
	index ^= 1;
	if (strncmp(sym_names[index], sym.a_name, 8) == 0)
	{
	    if ((found = (sym.a_type & A_DATA)) || all++)
		break;
	}
    }
    
    if(!found)
    {
	fprintf(stderr, "symbol _stksize is undefined\n");
	exit(9);
    }
    *what = index;
    return sym.a_value + sizeof(head);
}

int main(argc, argv)
int argc;
char **argv;
{
    int fd;
    int what;
    long stksize, offset;
    char fn[FILENAME_MAX];
    
    if(argc > 1)
	(void) strcpy(fn, *++argv);
    else
	(void) strcpy(fn, "gcc-cc1.ttp");
    
    if((fd = open(fn, 0)) < 0)
    {
	perror(fn);
	exit(1);
    }
    
    offset = find_offset(fd, fn, &what);
    
    if(lseek(fd, offset, 0) != offset)
    {
	perror(fn);
	exit(7);
    }
    read(fd, &stksize, sizeof(long));
    printf("%s: %s is %ld (%dK)\n",
    	 fn, sym_names[what] + 1, stksize, (int)(stksize/1024));
    
    return close(fd);
}

